#include "crogue.h"

// Inventory filters
int filter_none(item *i);
int filter_cursed(item *i);
int filter_worn(item *i);
int filter_wearable(item *i);
int filter_unidentified(item *i);
int filter_usable(item *i);

// Inventory menus
int pick_item(inventoryfilter f);
static void itemstring(int n, char *buf);
static int item_by_sequence(int n);
static void sort_inventory(void);
int filter_matches(inventoryfilter f);

// Player inventory
void init_inventory(void);
item split_item(int slot);
void merge_item(int slot, item it);
int give_item(const item* item, int silent);

// Equipment handling
item *player_equipment(int slot);
static int num_player_equipment(int slot);

// Utility functions
const char *shortitemname(const item *i);
uchar *itemname(const item *i);
static void append_armor_description(const item *itm, char *output);
static sint effect_amount(const stat_mod_desc *e, const item *itm);

// Misc
void list_discoveries(void);
void identify_type(int t);
ulong type_is_identified(ulong t);

// Item storage
void compress_items(void);
void place_item(item i, uchar x, uchar y);
uchar draw_item(uchar x, uchar y);
void init_items();
void cleanup_items(void);
uint add_item(item itm);
void delete_item(uint which);

item randitem(void);
static schar random_plus(void);

static uint can_stack(const item *one, const item *two);

// Player actions
void drop_item(void);
void wear_item(void);
void takeoff_item(void);
void pickup(void);

/*
 * Constant data
 */
//{{{
#define NUM_ITEM_CATEGORIES 10
const char *categoryNames[NUM_ITEM_CATEGORIES] =
	{
		"Weapons",
		"Ammunition",
		"Armor",
		"Rings",
		"Wands",
		"Potions",
		"Scrolls",
		"Spellbooks",
		"Food",
		"Gems"
	};
//}}}
//{{{
const size_t item_info_size[6] =
	{
		0,
		sizeof(weapon_info),
		sizeof(armor_info),
		sizeof(magic_info),
		sizeof(ammo_info),
		sizeof(spellbook_info)
	};
//}}}


/*
 * Inventory filters
 */
//{{{
int filter_none(item *i)
{
	return 1;
}
//}}}
//{{{
int filter_cursed(item *i)
{
	return (i->flags & ITEMFLAG_CURSED) && (i->flags & ITEMFLAG_IDENTIFIED);
}
//}}}
//{{{
int filter_worn(item *i)
{
	return i->flags & ITEMFLAG_EQUIPPED;
}
//}}}
//{{{
int filter_wearable(item *i)
{
	if(i->flags & ITEMFLAG_EQUIPPED)
		return 0;
	return ITEMDESC(*i).spot >= 0;
}
//}}}
//{{{
int filter_unidentified(item *i)
{
	if(!type_is_identified(i->type))
		return 1;
	else if(!(i->flags & ITEMFLAG_IDENTIFIED))
		return 1;
	else
		return 0;
}
//}}}
//{{{
int filter_usable(item *i)
{
	int type = ITEMDESC(*i).type;
	
	return (type==type_spellbook || type==type_magic);
}
//}}}

/*
 * Inventory menus
 */
static inventoryfilter invfilter;

//{{{
int pick_item(inventoryfilter f)
{
	int ret;
	int num_choices;
	
	if(w->plr.numitems==0)
		return -2;
	
	sort_inventory();
	
	invfilter = f;
	num_choices = filter_matches(f);
	
	if(!num_choices)
		return -2;
	
	ret = UI_Menu_Pick( COORD(7,14), COORD(142, 79), num_choices, &itemstring, 0);
	UI_TF_FullRedraw(&w->tilefield);
	
	if(ret<0)
		return -1;
	
	ret = item_by_sequence(ret);
	
	if( ret<0 )
		return -1;
	else
		return ret;
}
//}}}
//{{{
void itemstring(int n, char *buf)
{
	static int prev_n = -2;
	static int i = 0;
	static int prev_category;
	
	if(n == prev_n+1)
	{
		prev_n = n;
		n=1;
	} else {
		prev_n = n;
		prev_category = -1;
		i=0;
	}
	
	for(; i<PLRINVSIZE; i++)
	{
		if( !invfilter(&INVENTORY(i)) )
			continue;
		
		if(ITEMDESC(INVENTORY(i)).category != prev_category)
		{
			if(!n)
			{
				strcpy(buf, categoryNames[ITEMDESC(INVENTORY(i)).category]);
				break;
			}
			n--;
			
			prev_category = ITEMDESC(INVENTORY(i)).category;
		}
		
		if(!n)
		{
			sprintf(buf, "  %s", itemname(&INVENTORY(i)));
			break;
		}
		n--;
	}
}
//}}}
//{{{
// Works ONLY on a pre-sorted inventory. (use sort_inventory())
int item_by_sequence(int n)
{
	int i;
	int prev_category = -1;
	
	for(i=0; i<PLRINVSIZE; i++)
	{
		if( !invfilter(&INVENTORY(i)) )
			continue;
		
		if(ITEMDESC(INVENTORY(i)).category != prev_category)
		{
			if(!n)
				return -ITEMDESC(INVENTORY(i)).category - 1;
			n--;
			
			prev_category = ITEMDESC(INVENTORY(i)).category;
		}
		
		if(!n)
			return i;
		n--;
	}

	return 0;
}
//}}}
//{{{
void sort_inventory(void)
{
	int i, j;
	int num=0;
	int cont;
	item tmp;
	
	for(i=0; i<NUM_ITEM_CATEGORIES; i++)
		for(j=0; j<PLRINVSIZE; j++)
			if(INVENTORY(j).type && ITEMDESC(INVENTORY(j)).category == i)
				INVENTORY(j).next = num++;
	
	do
	{
		cont = 0;
		
		for(i=0; i<PLRINVSIZE; i++)
			if(INVENTORY(i).type && INVENTORY(i).next != i)
			{
				cont = 1;
				
				tmp = INVENTORY(INVENTORY(i).next);
				INVENTORY(INVENTORY(i).next) = INVENTORY(i);
				INVENTORY(i) = tmp;
			}
	}
	while(cont);
}
//}}}

//{{{
int filter_matches(inventoryfilter f)
{
	int i, j;
	int n=0;
	
	for(i=0; i<PLRINVSIZE; i++)
		if(INVENTORY(i).type && f(&INVENTORY(i)))
			n++;
	
	for(i=0; i<NUM_ITEM_CATEGORIES; i++)
		for(j=0; j<PLRINVSIZE; j++)
			if(INVENTORY(j).type && ITEMDESC(INVENTORY(j)).category == i && f(&INVENTORY(j)))
			{
				n++;
				break;
			}
	
	return n;
}
//}}}

/*
 * Player inventory
 */
//{{{
void init_inventory(void)
{
	const uint *ptr;
	const item *contents;
	uint i;
	uint num_items;
	
	// Initialize inventory empty
	for(i=0; i<PLRINVSIZE; i++)
		w->plr.inventory[i].type = 0;
	
	ptr = (const uint*)CONSTFILEDEREF(w->desc.startinginventory);
	num_items = *ptr;
	contents = (item*)(ptr+1);
	
	w->plr.numitems = num_items;
	
	for(i=0; i<num_items; i++)
	{
		INVENTORY(i) = contents[i];
		
		identify_type(INVENTORY(i).type);
	}
}
//}}}
//{{{
item split_item(int slot)
{
	item ret = INVENTORY(slot);
	
	ret.stacksize = 1;
	
	INVENTORY(slot).stacksize--;
	if(INVENTORY(slot).stacksize<=0)
	{
		INVENTORY(slot).type = 0;
		w->plr.numitems--;
	}
	
	return ret;
}
//}}}
//{{{
void merge_item(int slot, item it)
{
	if(INVENTORY(slot).type == 0) {
		INVENTORY(slot) = it;
	} else {
//		INVENTORY(slot).info.mag.charges += it.info.mag.charges;
		INVENTORY(slot).stacksize += it.stacksize;
	}
}
//}}}
//{{{
int give_item(const item* item, int silent)
{
	int i;
	
	// Look for items of the same type to stack with
	if(ITEMDESC(*item).flags & ITEMFLAG_STACK) {
		for(i=0; i<PLRINVSIZE; i++) {
			if( can_stack(&w->plr.inventory[i], item) )
			{
				w->plr.inventory[i].stacksize += item->stacksize;
				if( !silent ) {
					message("You have %s.", itemname(&w->plr.inventory[i]));
				}
				return 1;
			}
		}
	}
	// If it can't stack, look for empty slots
	for(i=0; i<PLRINVSIZE; i++) {
		if(!w->plr.inventory[i].type) {
			if( !silent ) {
				message("You got a %s.", itemname(item) );
			}
			w->plr.inventory[i] = *item;
			w->plr.numitems++;
			return 1;
		}
	}
	// Otherwise, give up
	message("You don't have room in your pack.");
	return 0;
}
//}}}

/*
 * Equipment handling
 */
//{{{
item *player_equipment(int slot)
{
	int i;
	
	for(i=0; i<PLRINVSIZE; i++)
	{
		if(INVENTORY(i).type && ITEMDESC(INVENTORY(i)).spot == slot && INVENTORY(i).flags & ITEMFLAG_EQUIPPED)
			return &INVENTORY(i);
	}
	
	return NULL;
}
//}}}
//{{{
int num_player_equipment(int slot)
{
	int i;
	int n=0;
	
	for(i=0; i<PLRINVSIZE; i++)
	{
		if(INVENTORY(i).type && ITEMDESC(INVENTORY(i)).spot == slot && INVENTORY(i).flags & ITEMFLAG_EQUIPPED)
			n++;
	}
	return n;
}
//}}}

/*
 * Utility functions
 */

//{{{
const char *shortitemname(const item *i)
{
	itemdesc *desc;
	desc = &ITEMDESC(*i);
	if( type_is_identified(i->type) || isNull(desc->unidname) )
		return (uchar*)CONSTFILEDEREF(desc->name);
	else
		return (uchar*)CONSTFILEDEREF(desc->unidname);
}
//}}}
//{{{
uchar *itemname(const item *i)
{
	const char *n;
	static char name[128];
	uint identified;
	itemdesc *desc;
	
	identified = i->flags & ITEMFLAG_IDENTIFIED;
	desc = &ITEMDESC(*i);
	
	n = shortitemname(i);
	
	strcpy(name, "");
	
	if(i->stacksize != 1 && desc->flags & ITEMFLAG_STACK)
		catprintf(name, "%i ", i->stacksize);
	if(i->flags & ITEMFLAG_CURSED && identified)
		catprintf(name, "cursed ");
	
	if(i->stacksize != 1 && desc->flags & ITEMFLAG_STACK)
		catprintf(name, pluralize(n));
	else
		catprintf(name, n);
	
	if(desc->flags & ITEMFLAG_CHARGED && identified)
		catprintf(name, " (%i charges)", i->stacksize);
	
	if(!identified)
		return name;
	
	switch( desc->type )
	{
		case type_weapon:
			catprintf(name, "(%i-%i,%i)",
				desc->desc.weapon.damage.min + i->info.weap.plus,
				desc->desc.weapon.damage.max + i->info.weap.plus,
				desc->desc.weapon.to_hit + i->info.weap.plus);
			if(i->flags & ITEMFLAG_EQUIPPED)
				catprintf(name, " (wielded)");
			break;
		case type_ammo:
			catprintf(name, "(%i-%i,%i)",
				desc->desc.ammo.damage.min + i->info.ammo.plus_dam,
				desc->desc.ammo.damage.max + i->info.ammo.plus_dam,
				desc->desc.ammo.to_hit     + i->info.ammo.plus_hit
				);
			if(i->flags & ITEMFLAG_EQUIPPED)
				catprintf(name, " (readied)");
			break;
		case type_armor:
			append_armor_description(i, name);
			if(i->flags & ITEMFLAG_EQUIPPED)
				catprintf(name, " (worn)");
			break;
	}
	return name;
}
//}}}
//{{{
void append_armor_description(const item *itm, char *output)
{
	int dv=0, pv=0, dmg_min=0, dmg_max=0, to_hit=0;
	int i;
	
	stat_mod_desc *effects = ITEMDESC(*itm).desc.armor.effects;
	int num_effects = ITEMDESC(*itm).desc.armor.num_effects;
	
	for(i=0; i<num_effects; i++)
	{
		if(effects[i].stat == STAT_DV)
			dv = effect_amount(&effects[i], itm);
		if(effects[i].stat == STAT_PV)
			pv = effect_amount(&effects[i], itm);
		if(effects[i].stat == STAT_DAMAGE_MIN)
			dmg_min = effect_amount(&effects[i], itm);
		if(effects[i].stat == STAT_DAMAGE_MAX)
			dmg_max = effect_amount(&effects[i], itm);
		if(effects[i].stat == STAT_TOHIT)
			to_hit = effect_amount(&effects[i], itm);
	}
	
	if(dv||pv)
	{
		catprintf(output, " [%i, %i]", dv, pv);
	}
	if(dmg_min || dmg_max)
	{
		catprintf(output, " (%i-%i)", dmg_min, dmg_max);
	}
	if(to_hit)
	{
		catprintf(output, " {%i}", to_hit);
	}
	
	for(i=0; i<num_effects; i++)
	{
		int amt = effect_amount(&effects[i], itm);
		
		switch(effects[i].stat)
		{
			case STAT_STRENGTH:
				catprintf(output, " (%i St)", amt);
				break;
			case STAT_DEXTERITY:
				catprintf(output, " (%i Dx)", amt);
				break;
			case STAT_TOUGHNESS:
				catprintf(output, " (%i To)", amt);
				break;
			case STAT_WILLPOWER:
				catprintf(output, " (%i Wi)", amt);
				break;
			case STAT_MANA:
				catprintf(output, " (%i Mn)", amt);
				break;
			case STAT_LIGHTRADIUS:
				catprintf(output, " (%i Light)", amt);
				break;
			case STAT_SPEED:
				catprintf(output, " (%i Spd)", amt);
				break;
		}
	}
}
//}}}
//{{{
sint effect_amount(const stat_mod_desc *e, const item *itm)
{
	if(e->type == STATMOD_ADD)
		return e->amount;
	else if(e->type == STATMOD_ADDPLUS)
		return e->amount + itm->info.armr.plus;
	
	return 0;
}
//}}}

/*
 * Misc
 */
const unsigned char discoveryScreenTabStops[3] = { 0, 70, 100 };
//{{{
void list_discoveries(void)
{
	int i;
	int num_discoveries = 0;
	char buf[256];
	draw_string_info state = {0, 0};
	
	setTabStops(3, discoveryScreenTabStops);
	
	for(i=0; i<w->desc.itementries; i++)
	{
		if(type_is_identified(i) && !isNull(ITEMDESCN(i).unidname))
		{
			if(!num_discoveries)
				Black_ClrScr();
			
			sprintf(buf, "%s\t%s\n", (const char*)CONSTFILEDEREF(ITEMDESCN(i).unidname),
			                       (const char*)CONSTFILEDEREF(ITEMDESCN(i).name));
			draw_string(buf, &state, 0, 0, SCREEN_COLUMNS, SCREEN_ROWS, 1);
			num_discoveries++;
		}
	}
	if(!num_discoveries)
	{
		message("You haven't identified anything interesting.");
		return;
	}
	READCHAR;
	UI_TF_FullRedraw(&w->tilefield);
}
//}}}
//{{{
void identify_type(int t)
{
	w->itemids[t/32] |= 1l << (t%32l);
}
//}}}
//{{{
ulong type_is_identified(ulong t)
{
	return w->itemids[t/32] & (1l << (t % 32));
}
//}}}

/*
 * Item storage
 */
const int items_startsize = 64;
const int items_allocinterval = 8;

//{{{
void compress_items(void)
{
	int i;
/*	// Do a pile compress
	int j, k;
	
	for(i=j=1; i<w->items.num; i++)
	{
		if(w->items.items[i].type)
		{
			memcpy(&w->items.items[j], &w->items.items[i], sizeof(item));
			
			// Re-target links
			for(k=0; k<w->items.num; k++)
				if(w->items.items[i].next == i)
					w->items.items[i].next = j;
			
			j++;
		}
	}
	
	// Zero out the end
	memset( &w->items.items[j], 0, w->items.alloced-j );*/
	
	// Zero out unused items
	for(i=0; i<w->items.num; i++)
	{
		if(!w->items.items[i].type)
			memset(&w->items.items[i], 0, sizeof(item));
	}
}
//}}}
//{{{
void place_item(item i, uchar x, uchar y)
{
	w->t[y][x].numitems++;
	i.next = w->t[y][x].topitem;
	w->t[y][x].topitem = add_item(i);
}
//}}}
//{{{
uchar draw_item(uchar x, uchar y)
{
	return ITEMDESC( w->items.items[w->t[y][x].topitem] ).drawchar;
}
//}}}
//{{{
void init_items()
{
	int i;
	
	w->items.num = 0;
	w->items.alloced = items_startsize;
	
	w->items.items = malloc(sizeof(item) * items_startsize);
	if(!w->items.items)
		fatal("Out of memory.");
	
	for(i=0; i<w->items.alloced; i++)
		w->items.items[i].type = 0;
}
//}}}
//{{{
void cleanup_items(void)
{
	if(w->items.items)
		free(w->items.items);
}
//}}}
//{{{
uint add_item(item itm)
{
	int i;
	
	// Allocate more memory if necessary
	w->items.num++;
	if(w->items.num >= w->items.alloced) {
		w->items.alloced += items_allocinterval;
		w->items.items = (item*)realloc(w->items.items, sizeof(item) * w->items.alloced);
		if(!w->items.items)
			fatal("Out of memory for items.");
		
		// Init newly created space
		for(i=w->items.alloced - items_allocinterval; i<w->items.alloced; i++) {
			w->items.items[i].type = 0;
		}
	}
	
	// Find an empty spot in the items array
	// i<items_alloced test unnecessary since resizing guarantees break first
	// Starting at 1 (wasting spot 0) to avoid confusing item 0 with 'no item'
	for(i=1; ; i++) {
		if(!w->items.items[i].type)
			break;
	}
	w->items.items[i] = itm;
	
	return i;
}
//}}}
//{{{
void delete_item(uint which)
{
	int i;

	// Find any items with link-fields pointing to the deleted item and redirect
	for(i=0; i<w->items.alloced; i++)
	{
		if(w->items.items[i].next == which)
			w->items.items[i].next = w->items.items[which].next;
	}
	w->items.items[which].type = 0;
	w->items.num--;
}
//}}}

//{{{
item randitem(void)
{
	item i;
	
	i.type = ((itemdesc*)CONSTFILEDEREF(w->desc.generate_item)) -> num;
	
	i.stacksize = 1;
	i.flags = 0;
	
	if(ITEMDESC(i).flags & (ITEMFLAG_STACK|ITEMFLAG_CHARGED))
		i.stacksize = RANGE(ITEMDESC(i).stacksize, ITEMDESC(i).stacksize/2);
	if(i.stacksize<1) i.stacksize = 1;
	
	switch( ITEMDESC(i).type )
	{
		case type_weapon:
			i.info.weap.plus = random_plus();
			if(i.info.weap.plus < 0)
				i.flags |= ITEMFLAG_CURSED;
			i.info.weap.special=0;
			break;
		case type_ammo:
			i.info.ammo.plus_dam = i.info.ammo.plus_hit = random_plus();
			break;
		case type_armor:
			i.info.armr.plus = random_plus();
			if(i.info.armr.plus < 0)
				i.flags |= ITEMFLAG_CURSED;
			i.info.armr.special=0;
			break;
		case type_spellbook:
			i.info.book.readings = 3;
			break;
	}
	
	return i;
}
//}}}
//{{{
schar random_plus(void)
{
//{{{
	const schar weighted_plus[16] = {
			1,	1,
			1,	1,
			1,	1,
			1,	1,	
			2,	2,	
			2,	2,	
			3,	3,	
			4,	5	
		};
//}}}
	int t;
	
	// One in 10 items is positively enchanted, another 1 in 10 is negatively enchanted.
	t = RANGE(10,1);
	if     (t==10) return  weighted_plus[rand()%16];
	else if(t==1)  return -weighted_plus[rand()%16];
	else           return  0;
}
//}}}

//{{{
uint can_stack(const item *one, const item *two)
{
	if(one->type != two->type)
		return 0;
	if( (one->flags ^ two->flags) & ~ITEMFLAG_EQUIPPED )
		return 0;
	
	if(item_info_size[ITEMDESC(*one).type]>0)
		if(memcmp(&one->info, &two->info, item_info_size[ITEMDESC(*one).type]))
			return 0;
	
	return 1;
}
//}}}

/*
 * Player actions
 */
//{{{
void drop_item(void)
{
	int which;
	
	if( filter_matches(filter_none) == 0 )
	{
		message("You don't have anything to drop.");
		return;
	}
	
	message("Drop what?");
	which = pick_item(filter_none);
	
	if(which<0) {
		message("Never mind.");
		return;
	}
	
	if(w->plr.inventory[which].flags & ITEMFLAG_EQUIPPED)
	{
		if( ITEMDESC(w->plr.inventory[which]).type == type_armor ) {
			message("You can't drop something you're wearing.");
			return;
		} else if( ITEMDESC(w->plr.inventory[which]).type == type_weapon ) {
			message("You need to unwield it first.");
			return;
		}
	}
	
	message("You drop the %s.", itemname(&w->plr.inventory[which]));
	place_item(w->plr.inventory[which], w->plr.x, w->plr.y);
	w->plr.inventory[which].type=0;
	w->plr.numitems--;
	
	timepass(1);
}
//}}}
//{{{
const int max_worn_items[9] =
	{
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		1,
		2
	};
//}}}
//{{{
void wear_item(void)
{
	int which, slot;
	item *unequip;
	
	if( filter_matches(filter_wearable) == 0 )
	{
		message("You don't have anything to equip.");
		return;
	}
	message("Equip what?");
	
	which=pick_item(filter_wearable);
	
	if(which<0) {
		message("Never mind.");
		return;
	}
	
	slot = ITEMDESC(INVENTORY(which)).spot;
	
/*	if(slot<0) // This is enforced by the menu filter
	{
		message("You can't wear that.");
		return;
	}*/
	
	if(num_player_equipment(slot) >= max_worn_items[slot])
	{
		unequip = player_equipment(slot);
		
		if( unequip )
		{
			if(unequip->flags & ITEMFLAG_CURSED)
			{
				message("You can't remove your old %s.", (slot)?("armor"):("weapon"));
				return;
			}
			unequip->flags &= ~ITEMFLAG_EQUIPPED;
			message("You %s the %s.", (slot>1)?("take off"):(slot==1)?("unready"):("unwield"), itemname(unequip));
		}
	}
	
	message( "You %s the %s.", (slot>1)?("put on"):(slot==1)?("ready"):("wield"), itemname(&w->plr.inventory[which]));
	
	INVENTORY(which).flags |= ITEMFLAG_EQUIPPED | ITEMFLAG_IDENTIFIED;
	identify_type(INVENTORY(which).type);
	
	update_player();
	
	timepass(1);
}
//}}}
//{{{
void takeoff_item(void)
{
	int which, slot;
	
	if( filter_matches(filter_worn) == 0 )
	{
		message("You don't have anything to unequip.");
		return;
	}
	
	message("Unequip what?");
	which = pick_item(filter_worn);
	
	if(which<0)
	{
		message("Never mind.");
		return;
	}
	
	slot = ITEMDESC(w->plr.inventory[which]).spot;
	
	if( !(INVENTORY(which).flags & ITEMFLAG_EQUIPPED) )
	{
		message("You aren't wearing/wielding that.");
		return;
	}
	
	if( INVENTORY(which).flags & ITEMFLAG_CURSED )
	{
		message("You can't. It seems to be cursed.");
		return;
	}

	w->plr.inventory[which].flags &= ~ITEMFLAG_EQUIPPED;
	
	message( "You %s the %s.", (slot>1)?("take off"):(slot==1)?("unready"):("unwield"), itemname(&w->plr.inventory[which]));
	
	update_player();
	
	timepass(1);
}
//}}}
//{{{
void pickup(void)
{
	sshort x=w->plr.x, y=w->plr.y;
	uint pos, next;
	
	pos = w->t[y][x].topitem;
	
	while(pos)
	{
		next = w->items.items[pos].next;
		message("Pick up the");
		if( prompt(itemname(&w->items.items[pos])) )
		{
			if(give_item(&w->items.items[pos], 0))
			{
				if(pos==w->t[y][x].topitem)
					w->t[y][x].topitem = next;
				delete_item(pos);
				w->t[y][x].numitems--;
				timepass(1);
			}
		}
		pos = next;
	}
}
//}}}

